/*this file modified by Tom Pratum, July 2007*/
/*derived from files originally in SpinCore spinapi dated 2-28-2007, modified based on spinapi version from 7/30/2007*/
/*8/29/08 usb_write_reg and usb_read_reg changed to correct for byte order */
/*2/01/2009 os_usb_readTO changed to correctly take into account timeouts - note that if there is a timeout, generally the data need to be read again, even though the function deosn't automatically do that now */
/* Jan, Feb 2011: updated to latest spinapi, and made reads asynchronous for getting data */

#include <unistd.h>

#include <CoreFoundation/CoreFoundation.h>

#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/usb/IOUSBLib.h>
/*#include <Kernel/IOKit/usb/IOUSBLog.h>*/

#include <mach/mach.h>


/*#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>*/
#include "if.h"
#include "usb.h"
#include "util.h"
#include "caps.h"
#include "spinapi.h"

// The USB device can accept at most 512 bytes in a single transfer
#define MAX_XFER_SIZE 512

/*global definition of usb interface; note definition for 10.5 and above*/
extern IOUSBInterfaceInterface300 	**intfg;

extern char *spinerr;

extern char *noerr;

/*the values of these two should be set in do_os_init*/
extern int cur_board;
extern int num_boards;

/* array to hold pid values */
extern int pid_list[MAX_NUM_BOARDS];

// This array holds the capabilties info on each board
/* note that you can have up to 8 boards due to the way os_init is written */
BOARD_INFO board[MAX_NUM_BOARDS];

int setup_xfer(unsigned int addr, unsigned int packet_len);

/*int pb_outp(unsigned int address, char data);*/

/*int pb_outp(unsigned int address, unsigned int data);*/

extern int cur_dev;

//default portbase supplied to backwards compatibility
static int port_base = 0; 


/*this function differs from the spinapi function - it initializes ALL of the boards at once rather than only the board given as an argument
Otherwise, it is the same*/
int pb_init()
{
    spinerr = noerr;

    int dac_control;
    int adc_control;

	int dev_id, i;

   	/*this statement initializes all of the boards wrt the os*/
   	dev_id = do_os_init();

    if (dev_id == -1) {
        // spinerr should have been set inside the os_init
		debug ("pb_init error (os_init failed): %s\n", spinerr);
        return -1;
    }

    // Figure out what the capabilities of all of the boards are
	for(i=0; i<num_boards; i++)
	{
	 if(!pb_select_board(i))
	 	{
	 	 debug("cannot select board %d in pb_init!\n",i);
	 	 return -1;
	 	 }
	 cur_board = i;
	 /*printf("getting capabilities of board %d, dev id %d, 0x%x, 0x%x\n",cur_board,dev_id,dev_id,pid_list[cur_board]);*/
	 if (get_caps (&(board[cur_board]), pid_list[cur_board]) < 0)
	 {
		// spinerr should have been set inside get_caps
		debug ("pb_init error (get_caps failed): %s\n", spinerr);
		return -1;
		}
     // If this is a RadioProcessor, set ADC and DAC defaults. This is done here instead of pb_set_defaults() because the user should not ever change these values.
     if(board[cur_board].is_radioprocessor)
     {
    	debug("pb_init: this is a radioprocessor board\n");
    	adc_control = 3;
        dac_control = 0;
        pb_set_radio_hw(adc_control, dac_control);
        }
      board[cur_board].did_init = 1;
      }
    /*reset current board to zero*/
    if(!pb_select_board(0))
    	{
    	 debug("cannot select board 0 in pb_init\n");
    	 return -1;
    	 }

    return 0;
}

int pb_close()
{
    spinerr = noerr;

    if(board[cur_board].did_init == 0) {
        spinerr = "Board is already closed";
        debug("pb_close: %s\n", spinerr);
        return -1;
    }

    board[cur_board].did_init = 0;
    return os_usb_close();
}


int os_usb_close()
{
    return 0;   
    
}

int do_os_init()
{
    int dev_id, i;

        debug("do_os_init: initializing usb\n");
        dev_id = os_usb_init();        
        if(dev_id != -1)
        	{
        	 for(i = 0; i<num_boards; i++)
        	 	{
        	 	 /*printf("selecting board %d in do_os_init\n",i);*/
        	 	 if(!pb_select_board(i))
        	 	 {
        	 	  debug("cannot select board %d in do_os_init!\n",i);
        	 	  return -1;
        	 	  }
        	 	 usb_reset_gpif();
        	 	 }
        	 return dev_id;
        	 }
        else
        	return -1;
}

int usb_reset_gpif()
{
    char buf[3];

    buf[0] = RST_L | DO_GPIF_RESET | 0x01;
    os_usb_write(EP1OUT,buf,1);

    return 0;
}

/*several of the functions below differ from their spinapi counterparts only by the variable type for data */

int pb_outp(unsigned int address, unsigned int data)
{
    spinerr = noerr;

    debug2("\tpb_outp: addr %x, data %x. \n", address, data);

	/*this only works for usb devices*/
	
	return usb_do_outp(address, data);
}

int usb_do_outp(unsigned int address, unsigned int data)
{
    unsigned int word;
    
    word = 0x0FF & data;
    word = word | ((0x07 & address) << 8);
    unsigned int load_flag = 0x01 << 11;

    if(usb_write_reg(REG_PBCORE, word) < 0) {
      debug("pb_outp_usb: error 1\n");
      return -1;                    
    }
    if(usb_write_reg(REG_PBCORE, word | load_flag) < 0) {
      debug("pb_outp_usb: error 2\n");
      return -1;                    
    }

    return 0;
}

void os_usb_reset_pipes()
{
	UInt8 i;
    IOReturn			kr;

	for(i=1; i<=3; i++)
		{
			kr = (*intfg)->AbortPipe(intfg, i);
     		if (kIOReturnSuccess != kr)
        	{
            	debug("unable to do pipe abort on pipe %d\n", i);
            	}
			kr = (*intfg)->ResetPipe(intfg, i);
     		if (kIOReturnSuccess != kr)
        	{
            	debug("unable to do pipe reset on pipe %d\n", i);
            	}
			kr = (*intfg)->ClearPipeStall(intfg, i);
     		if (kIOReturnSuccess != kr)
        	{
            	debug("unable to do pipe stall clear on pipe %d\n", i);
            	}
			kr = (*intfg)->ClearPipeStallBothEnds(intfg, i);
     		if (kIOReturnSuccess != kr)
        	{
            	debug("unable to do both ends pipe stall clear on pipe %d\n", i);
            	}
          }
}

int usb_write_reg(unsigned int addr, unsigned int data)
{
    int ret;

    os_usb_reset_pipes();

    ret = setup_xfer(addr, 4);
    if (ret < 0) {
        spinerr = "Error setting up transfer";
        debug("usb_write_reg: %s\n",spinerr);
        return ret;
    }
// Make sure the byte order is correct
	data = CFSwapInt32HostToLittle(data);
	    
    ret = os_usb_write(EP2OUT, &data,4);
    if (ret < 0) {
        spinerr = "Error doing write";
        debug("usb_write_reg: %s\n",spinerr);
        return ret;
    }

    return 0;
}

int usb_read_reg(unsigned int addr, unsigned int *data)
{
    int ret, i, nPreReads;
    UInt32 size;

    /*seems like the number of pre-reads should be three, rather than two as the original software showed - subsequent reads all return the same thing*/
    nPreReads = 3;
    os_usb_reset_pipes();

    // Original text follows: read 3 times. the first two will bring back whatever
    // was in the buffers because the chip is double buffered
    // the third read will have the correct value

    for(i= 0; i < (nPreReads + 1); i++) {
        ret = setup_xfer(addr, 4);
        if (ret < 0) {
            spinerr = "Error setting up transfer";
            debug("usb_read_reg: %s (i=%d)\n",spinerr, i);
            return ret;
        }

        size = sizeof(data);
        ret = os_usb_read(EP6IN, data, &size);
        /*printf("cur_board is %d, i is %d, data is %d, 0x%x\n",cur_board,i,*data,*data);*/
        if (ret < 0) {
            spinerr = "Error doing read";
            debug("usb_read_reg: %s (i=%d)\n",spinerr,i);
            return ret;
        }
    }

// Make sure the byte order is correct
	 *data = CFSwapInt32LittleToHost(*data);

    return 0;
}

int setup_xfer (unsigned int addr, unsigned int packet_len)
{
  char buf[3];
  char addr_buf[4];

  addr_buf[3] = (addr >> 24) & 0x0FF;
  addr_buf[2] = (addr >> 16) & 0x0FF;
  addr_buf[1] = (addr >> 8) & 0x0FF;
  addr_buf[0] = addr & 0x0FF;

  buf[2] = (packet_len >> 8) & 0x0FF;
  buf[1] = packet_len & 0x0FF;

  // set the xfer counter (which is only used for IN xfers)
  buf[0] = ADDR_REG_EN | RST_L | 0x01;
  os_usb_write (EP1OUT, buf, 3);

  // write the actual address
  if (pid_list[cur_dev] == 0xc2a9)
    {
      os_usb_write (EP2OUT, addr_buf, 4);
    }
  else
    os_usb_write (EP2OUT, addr_buf, 2);
  // now disable the address register (and thus enable
  // data registers)
  buf[0] = RST_L | 0x01;

  os_usb_write (EP1OUT, buf, 3);

  return 0;
}


/* in the next two functions note that there is a difference in how the size argument is passed in the OSX usb interface*/

int os_usb_write(int pipe, void *data, int size)
{
    IOReturn			kr;
    UInt8			intfNumEndpoints;
    UInt16			prodID;

    /*if interface is invalid, possibly due to previous error, then exit with error */
    kr = (*intfg)->GetNumEndpoints(intfg, &intfNumEndpoints);
    if (kIOReturnSuccess != kr)
        {
            debug("os_usb_write: invalid interface; unable to get number of endpoints (%08x)\n", kr);
			return -1;
			}
     kr = (*intfg)->GetDeviceProduct(intfg,&prodID);
    if (kIOReturnSuccess != kr)
        {
            debug("os_usb_read: unable to get product id (%08x)\n", kr);
			}
     debug2("entering os_usb_write board %d pipe %d, size %d, product id 0x%x; ret 0x%x\n",cur_board,pipe,size,prodID,kr);

    /*note that the current device was stripped out - this assumes only one device on the usb bus*/
    debug2("entering os_usb_write pipe %d, size %d\n",pipe,size);
    if(size > MAX_XFER_SIZE) {
        spinerr = "Transfer size is too big";
        debug("os_usb_write: %s\n", spinerr);
        return -1;
    	}
	/* check for non-existent interface pointer here */
	
     kr = (*intfg)->WritePipe(intfg, (UInt8)pipe, data, (UInt32)size);
     if (kIOReturnSuccess != kr)
        {
            debug("unable to do bulk write (%08x)\n", kr);
            (void) (*intfg)->USBInterfaceClose(intfg);
            (void) (*intfg)->Release(intfg);
            return -1;
        }
    return  0;
}

int os_usb_read(int pipe, void *data, UInt32 *size)
{
    IOReturn			kr;
    UInt8			intfNumEndpoints;
    UInt16			prodID;

    /*if interface is invalid, possibly due to previous error, then exit with error */
    kr = (*intfg)->GetNumEndpoints(intfg, &intfNumEndpoints);
    if (kIOReturnSuccess != kr)
        {
            debug("os_usb_read: invalid interface; unable to get number of endpoints (%08x)\n", kr);
			return -1;
			}

     kr = (*intfg)->GetDeviceProduct(intfg,&prodID);
    if (kIOReturnSuccess != kr)
        {
            debug("os_usb_read: unable to get product id (%08x)\n", kr);
			}
     debug2("entering os_usb_read pipe %d, size %d, product id 0x%x; ret 0x%x\n",pipe,*size,prodID,kr);
     if(*size > MAX_XFER_SIZE) {
        spinerr = "Transfer size is too big";
        debug("os_usb_read: %s\n", spinerr);
        return -1;
    	}

     kr = (*intfg)->ReadPipe(intfg, (UInt8)pipe, data, size);
     if (kIOReturnSuccess != kr)
        {
            debug("unable to do bulk read (0x%08x ,overrun: 0x%08x)\n", kr,kIOReturnOverrun);
            /*(void) (*intfg)->USBInterfaceClose(intfg);
            (void) (*intfg)->Release(intfg);*/
            return -1;
        }
      return 0;

}

int os_usb_readTO(int pipe, void *data, UInt32 *size)
{
    /* this function differs from that above only in that it checks for timeouts*/
    IOReturn			kr, kr2;
    /* set timeouts to 10 msec for no data, 250 msec for completion of read*/
    UInt8			intfNumEndpoints;
    UInt32 noDataTimeout = 1;
    UInt32 completionTimeout = 1, tmp_size;
	/* check for non-existent interface pointer here */
	
    /*if interface is invalid, possibly due to previous error, then exit with error */
    kr = (*intfg)->GetNumEndpoints(intfg, &intfNumEndpoints);
    if (kIOReturnSuccess != kr)
        {
            debug("os_usb_readTO: invalid interface; unable to get number of endpoints (%08x)\n", kr);
			return -1;
			}

     debug2("entering os_usb_read pipe TO %d, size %d\n",pipe,*size);
     tmp_size = *size;
     if(*size > MAX_XFER_SIZE) {
        spinerr = "Transfer size is too big";
        /* test stmt */
        debug("os_usb_read: %s\n", spinerr);
        return -1;
    	}
 
     kr = (*intfg)->ReadPipeTO(intfg, (UInt8)pipe, data, size, noDataTimeout, completionTimeout);
     if(tmp_size != *size)
     	printf("number of bytes returned does not match; returned: %d vs %d requested\n",*size,tmp_size);
     if (kIOReturnSuccess != kr)
        {
            /*test stmt */
            /*printf("unable to do bulk read, may have timed out (0x%08x, timeout: 0x%08x, overrun: 0x%08x)\n", kr,kIOReturnTimeout,kIOReturnOverrun);*/
			kr2 = (*intfg)->GetPipeStatus(intfg, (UInt8)pipe);
			printf("unable to do bulk read, may have timed out (0x%08x, timeout: 0x%08x, overrun: 0x%08x, pipe status 0x%08x), size on return %d\n", kr,kIOReturnTimeout,kIOReturnOverrun,kr2,*size);
			debug("unable to do bulk read, may have timed out (0x%08x, timeout: 0x%08x, overrun: 0x%08x)\n", kr,kIOReturnTimeout,kIOReturnOverrun);
			/* test stmt */
			/*printf("clearing pipe stall\n");*/
			kr = (*intfg)->ClearPipeStallBothEnds(intfg, (UInt8)pipe);
     		if (kIOReturnSuccess != kr)
        	{
            	/* test stmt */
            	/*printf("os_usb_readTO: unable to do both ends pipe stall clear on pipe %d\n", pipe);*/
            	debug("os_usb_readTO: unable to do both ends pipe stall clear on pipe %d\n", pipe);
            	}            
            /* don't close interface, instead try another read
            (void) (*intfg)->USBInterfaceClose(intfg);
            (void) (*intfg)->Release(intfg);
            */
     		/* try one more read after clearing pipe stall, if failure then return */
     		kr = (*intfg)->ReadPipeTO(intfg, (UInt8)pipe, data, size, noDataTimeout, completionTimeout);
     		if (kIOReturnSuccess != kr)
            		{
            			kr2 = (*intfg)->GetPipeStatus(intfg, (UInt8)pipe);
            			printf("second read after timeout failed - read attemp returned as failed, pipe status 0x%08x\n",kr2);
            			debug("second read after timeout also failed\n");
            			return -1;
            			}
        }
      return 0;

}

/*function to execute when asynchronous read completes, see function below */
void ReadCompletion(void *refCon,IOReturn result,void *arg0)
{
    /* IOUSBInterfaceInterface **interface =
	(IOUSBInterfaceInterface **) refCon;
    UInt32      numBytesRead = (UInt32) arg0;*/
 	CFRunLoopRef runLoop = CFRunLoopGetCurrent();
 	
    /*printf("Asynchronous bulk read complete, status =(%08x)\n", result); fflush(stdout);*/
    if (result != kIOReturnSuccess) 
    {
 		printf("Asynchronous bulk read failed\n");
        return;
    }
    /* this stops the loop after completion */
    CFRunLoopStop(runLoop);
    return;
}

/* asynchronous read function */
int os_usb_readAsync(int pipe, void *data, UInt32 *size)
{
    /* this function differs from that above only in that it checks for timeouts*/
    IOReturn			kr, kr2;
    CFRunLoopSourceRef          runLoopSource;
    CFRunLoopRef runLoop = CFRunLoopGetCurrent();
    IOUSBInterfaceInterface300     **interface = NULL;
    /* set timeouts to 10 msec for no data, 250 msec for completion of read*/
    UInt8			intfNumEndpoints;
    UInt32 tmp_size;
    /*this is the delay prior to running the runloop - making it larger can make the transfers more reliable */
    double read_delay = 0.01;
    
	/* check for non-existent interface pointer here */
	
    /*if interface is invalid, possibly due to previous error, then exit with error */
    kr = (*intfg)->GetNumEndpoints(intfg, &intfNumEndpoints);
    if (kIOReturnSuccess != kr)
        {
            debug("os_usb_readAsync: invalid interface; unable to get number of endpoints (%08x)\n", kr);
			return -1;
			}

     debug2("entering os_usb_read pipe Async %d, size %d\n",pipe,*size);
     tmp_size = *size;
     if(*size > MAX_XFER_SIZE) {
        spinerr = "Transfer size is too big";
        /* test stmt */
        debug("os_usb_read: %s\n", spinerr);
        return -1;
    	}
 
     (*intfg)->CreateInterfaceAsyncEventSource(intfg,&runLoopSource);
        if (kr != kIOReturnSuccess)
        {
            printf("Unable to create asynchronous event source (%08x)\n", kr);
        }
        CFRunLoopAddSource(runLoop,runLoopSource, kCFRunLoopDefaultMode);
        /*printf("About to do readpipeAsync\n");*/
        kr = (*intfg)->ReadPipeAsync(intfg,(UInt8)pipe,data,*size,ReadCompletion,(void *)intfg);
        pb_sleep_ms_float(read_delay);
		/*give read 300 usec to finish, then reset */
		CFRunLoopRun();
		/*kr2 = (*intfg)->GetPipeStatus(intfg, (UInt8)pipe);
		printf("pipe status before pipestallclear 0x%08x\n",kr2);*/      
		kr2 = (*intfg)->ClearPipeStallBothEnds(intfg, (UInt8)pipe);
		/*kr2 = (*intfg)->GetPipeStatus(intfg, (UInt8)pipe);
		printf("pipe status after pipestallclear 0x%08x\n",kr2);*/      
		if (kr!= kIOReturnSuccess)
        {
            printf("Unable to perform asynchronous bulk read (%08x %d)\n", kr, kr);
        }
     if (kIOReturnSuccess != kr)
        {
			kr2 = (*intfg)->GetPipeStatus(intfg, (UInt8)pipe);
			printf("unable to do bulk read, may have timed out (0x%08x, timeout: 0x%08x, overrun: 0x%08x, pipe status 0x%08x), size on return %d\n", kr,kIOReturnTimeout,kIOReturnOverrun,kr2,*size);
			debug("unable to do bulk read, may have timed out (0x%08x, timeout: 0x%08x, overrun: 0x%08x)\n", kr,kIOReturnTimeout,kIOReturnOverrun);
			/* test stmt */
			/*printf("clearing pipe stall\n");*/
            /* don't close interface, instead try another read
            (void) (*intfg)->USBInterfaceClose(intfg);
            (void) (*intfg)->Release(intfg);
            */
     		/* try one more read after clearing pipe stall, if failure then return */
        	kr = (*intfg)->ReadPipeAsync(intfg,(UInt8)pipe,data,*size,ReadCompletion,(void *)intfg);
        	pb_sleep_ms_float(read_delay);
			/*give read 300 usec to finish, then reset */
			CFRunLoopRun();
     		if (kIOReturnSuccess != kr)
            		{
            			kr2 = (*intfg)->GetPipeStatus(intfg, (UInt8)pipe);
            			printf("second read after timeout failed - read attemp returned as failed, pipe status 0x%08x\n",kr2);
            			debug("second read after timeout also failed\n");
            			return -1;
            			}
        }
      return 0;

}


 /**
 * \internal
 * Write a 32 bit word to the extended register given by address.
 *
 */
void reg_write (unsigned int address, unsigned int data)
{
        int return_value;
        
        return_value = usb_write_reg(address, data);
}

/**
 * \internal
 * Read a 32 bit word from the extended register given by address.
 *
 */
unsigned int reg_read (unsigned int address)
{
    int return_value;
    /* note that ret is the data returned */
    unsigned int ret;
    
    return_value = usb_read_reg(address, &ret);
    return ret;
}

/**
 * \internal
 * 
 * \bank Current choices are BANK_DATARAM (for the acquisition memory) and BANK_DDSRAM (for the shape waveforms)
 * \start_addr the starting address to write to. Note that this is the offset IN TERMS OF THE MEMORY WIDTH, and
 * not the offset in bytes. For example for the dataram memory with a width=8 bytes, address 0 is bytes 0-7, address
 * 1 is bytes 8-15, etc.
 * \param length (in bytes) of data to write to a RAM bank
 * \param data A buffer of the length you wish to write
 */

int ram_write(unsigned int bank, unsigned int start_addr, unsigned int len, char *data)
{
    int i;

    spinerr = noerr;

    int data_word;
    int write_flag = 0x0100;

    if(board[cur_board].is_usb) {
        return usb_write_ram(bank, start_addr, len, data); 
    }
    else {
        if(bank == BANK_DDSRAM) {
            debug("ram_write: Writing RAM with PCI method\n");
            // For each byte, write the data. Then set the write flag
            for(i = 0; i < len; i++) {
                data_word = 0x0FF & ((int)data[i]);
                reg_write(0x17, data_word);
                reg_write(0x17, data_word | write_flag);                
                reg_write(0x17, data_word);
            }
            return 0;
        }
        else {
            // Note: see pb_get_data() for how to read the acquisition memory on PCI boards
            spinerr="Internal error: writing to RAMS other than BANK_DDSRAM not supported on PCI";
            debug("ram_write: %s\n", spinerr);
            return -1;
        }
        
    }
    
}


/**
 * 
  function used above * 
 */

int usb_write_ram(unsigned int bank, unsigned int start_addr, unsigned int len, char *data)
{
    char *outbuf;
    int i;
    char *ptr;

    int num_xfers;
    int xfer_size;
    int line_size;
    int excess_xfer;

    switch(bank) {
        case BANK_DATARAM:
            line_size = 8;
            xfer_size = 512;
            usb_write_reg(0x0012, start_addr);
            break;
        
        case BANK_DDSRAM:
            line_size = 1;
            xfer_size = 512;
            break;
        
        default:
            debug("usb_write_ram: invalid bank (0x%x)\n", bank);
            return -1;
    }


    num_xfers = len / xfer_size;
    excess_xfer = len - num_xfers*xfer_size;

    if(len % line_size != 0) {
        debug("usb_write_ram: length is not multiple of line size\n");
        return -1;
    }

    outbuf = (char *)malloc(xfer_size);
    if(!outbuf) {
        debug("usb_write_ram: couldnt allocate scratchpad memory\n");
        return -1;
    }

    ptr = data;

    // setup transfer to work on pbram
    setup_xfer(bank, xfer_size);
   
    for(i=0;i<num_xfers;i++) {

        memcpy(outbuf, ptr, xfer_size);

        if(os_usb_write(EP2OUT, outbuf, xfer_size) < 0) {
            debug("write not succesfful (xfer %d)\n", i);
            return -1;        
        }

        ptr += xfer_size;
    }    

    if(excess_xfer != 0) {
        memcpy(outbuf, ptr, excess_xfer);
        if(os_usb_write(EP2OUT, outbuf, excess_xfer) < 0) {
            debug("write not succesfful (excess xfer)\n");
            return -1;        
        }
    }


    free(outbuf);

    return 0;
}


int my_usb_read_ram(unsigned int bank, unsigned int start_addr, unsigned int len, char *data);

/**
 * \internal
 * See documentation for ram_read() in if.c
 * 
 */
int usb_read_ram(unsigned int bank, unsigned int start_addr, unsigned int len, char *data)
{
    
    int ret = 0;
    
    /*printf("usb_read_ram: looking for %d bytes total\n",len);*/
    debug("usb_read_ram: looking for %d bytes total\n",len);
    if(bank == BANK_DATARAM) {
        // If this is the dataram bank, do the transfer as a series of smaller transfers. This affects
        // the transfer rate somewhat, but it reduces the chance we will have problems with transfer
        // timeouts

        // do transfer in 64kbyte chunks (which corresponts to 8k complex points or 32k direct points)
        // address offset is divided by 8, because DATARAM port is 64bits wide
        int my_len = 64*1024;
        int i;

        // FIXME: with some additional code we could remove this limitation
        if(len % my_len != 0) {
            debug("DATARAM transfers MUST be a multiple of 64kbytes (for now)\n");
            return -1;
        }
        for(i=0; i < len / my_len; i++) {
            /*printf("entering my_usb_read_ram, xfer %d\n",i);*/
            ret = my_usb_read_ram(bank, start_addr + my_len*i/8, my_len, data + my_len*i);
            if(ret < 0) break;
        }
    }
    else {
        ret = my_usb_read_ram(bank, start_addr, len, data);
    }

    return ret;
}

/*** the function below was modified to use the IOUSBlib functions that respond to timeouts; it looks a bit different from the spinapi version ***/
/**
 * \internal 
 * This function actually does the ram reads. Before it starts reading, it starts the watchdog timer.
 * Then after every call to os_usb_read(), we check to see if the watchdog timer ran out. If it has,
 * that means that the transfer timed out, and we must restart the entire transfer. The watchdog
 * timer code will reset the endpoints automatically on a timeout. FYI, if the transfer times out
 * and we DONT have a way to automaitally reset it, the only way the use can recover is by cycling
 * the power on their board.
 * 
 * (note: I am not sure why these timeouts occur, and whether it is a problem with our firmware, with
 * spinapi, with the Cypress drivers, or something else)
 * 
 */
int my_usb_read_ram(unsigned int bank, unsigned int start_addr, unsigned int len, char *data)
{
    char inbuf[512];
    int i;

    char *ptr;

    char buf[3];
    
    int num_xfers, excess_xfer, line_size;
    UInt32 xfer_size, xfer_size_tmp;

    int amount_xferred = 0, ret = 0;

    debug("my_usb_read_ram: looking for %d points\n",len);
    switch(bank) {
        case BANK_DATARAM:
            line_size = 8;
            xfer_size = 512;
            usb_write_reg(0x0012, start_addr);
            break;
        case BANK_DDSRAM:
            debug("usb_read_ram: DDRSRAM is write only\n");
            return -1;
            break;
        default:
            debug("usb_read_ram: invalid RAM bank\n");
            return -1;
            break;
    }

    num_xfers = (len+line_size) / xfer_size;
    excess_xfer = (len+line_size) - num_xfers*xfer_size;

    if(len % line_size != 0) {
        debug("usb_read_ram: length is not multiple of line size\n");
        return -1;
    }

    buf[2] = (xfer_size >> 8) & 0x0FF;
    buf[1] = xfer_size & 0x0FF;
    buf[0] = RST_L | DO_LITE;
   
    ptr = data;

    // setup transfer to work on dataram
    setup_xfer(bank, xfer_size);

    for(i=0;i<num_xfers+2;i++) {
        os_usb_write(EP1OUT, buf, 1); // no reset, address register enable is disabled
        /* not sure why I have to do this here (make a temp variable), but the size seems to initially be corrupted */
        xfer_size_tmp = xfer_size;
        /*printf("transfer number %d, bytes total %d\n", i, i*xfer_size_tmp);*/
        debug("transfer number %d, bytes total %d\n", i, i*xfer_size_tmp);
        /* test stmts: */
        /* printf(".%d.",i); fflush(stdout); */
        if(os_usb_readAsync(EP6IN, inbuf, &xfer_size_tmp) < 0) {
            /* test stmt */
            /*printf("usb_read_ram: read not successful (xfer %d)\n", i);*/
            debug("usb_read_ram: read not successful (xfer %d)\n", i);
            return -1;        
        }
        
        if(i==0 || i==1) {
            continue;
        }

        // on the first real transfer, skip the first line
        if(i==2) {
            memcpy(ptr, inbuf+line_size, xfer_size-line_size);
            ptr += xfer_size-line_size;
        }
        else {
            memcpy(ptr, inbuf, xfer_size);
            ptr += xfer_size;
        }

    }

    // if there is excess left to read, read an entire xfer_size block, but only
    // copy the part we want
    if(excess_xfer != 0) {
        debug("usb_read_ram: reading excess %d\n",xfer_size);
        os_usb_write(EP1OUT, buf, 1); // no reset, address register enable is disabled
        ret = os_usb_readAsync(EP6IN, inbuf, &xfer_size);
        if(ret < 0) {
            /* test stmt */
            /*printf("usb_read_ram: read not successful (excess)\n");*/
            debug("usb_read_ram: read not successful (excess)\n");
            return -1;        
        }
        
        memcpy(ptr, inbuf, excess_xfer);
    }


    // read two more times to clear out the buffer
    debug("usb_read_ram: clearing buffer %d\n",xfer_size);
    if(os_usb_readAsync(EP6IN, inbuf, &xfer_size) < 0) {
         /* test stmt */
         /*printf("usb_read_ram: read not succesful (clear 1)\n");*/
         debug("usb_read_ram: read not succesful (clear 1)\n");
         return -1;        
    }
    /* when read hangs, it is always on the second clear, however, it doesn't matter if we put in the second clear-read or not, the data will still at times be corrupted if there is an insufficient delay between reads above*/
    if(os_usb_readAsync(EP6IN, inbuf, &xfer_size) < 0) {
        /* test stmt */
        /*printf("usb_read_ram: read not succesful (clear 2)\n");*/
        debug("usb_read_ram: read not succesful (clear 2)\n");
        /* an error is only given if the 1st clear fails */
        return 0;       
    }


    os_usb_reset_pipes();

    reg_read(REG_CONTROL);
    reg_read(REG_CONTROL);

    /* note that this just returns 0 since they don't keep track of it */
    return amount_xferred;
}

/**
 * This function allows you to pause execution of your software for a given
 * number of milliseconds, and behaves like the sleep() function found on
 * many operating systems. This is provided because the sleep() function is not
 * standard across all operating systems/programming environments.
 * <p>This function does *NOT* interact with the hardware or pulse program at all. It
 * simply provides a portable way to access a sleep function.
 *
 * \param milliseconds Number of milliseconds to sleep (pause program) for.
 *
 */
 /* note that it was modified to use the nanosleep function in order to prevent compilation errors for osx 10.5 targets */

void pb_sleep_ms(int milliseconds)
{  
    #include <time.h>

	struct timespec req={0}, req_ret={0};  
	time_t sec=(int)(milliseconds/1000);  
	milliseconds=milliseconds-(sec*1000);  
	req.tv_sec=sec;  
	req.tv_nsec=milliseconds*1000000L;  
	nanosleep(&req,&req_ret); 

}

/* this function is the same as the above, but allows delays of fractions of msec
it is only used in the usb pipe reading functions */
void pb_sleep_ms_float(double milliseconds)
{  
    #include <time.h>

	struct timespec req={0}, req_ret={0};  
	time_t sec=(int)(milliseconds/1000);  
	milliseconds=milliseconds-(sec*1000);  
	req.tv_sec=sec;  
	req.tv_nsec=milliseconds*1000000L;  
	nanosleep(&req,&req_ret); 

}


